##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
#############################################################################

import unittest

from OFS.SimpleItem import SimpleItem
from Testing.makerequest import makerequest
from BTrees.IIBTree import difference


class UnIndexTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.PluginIndexes.common.UnIndex import UnIndex
        return UnIndex

    def _makeOne(self, *args, **kw):
        index = self._getTargetClass()(*args, **kw)

        class DummyZCatalog(SimpleItem):
            id = 'DummyZCatalog'

        # Build pseudo catalog and REQUEST environment
        catalog = makerequest(DummyZCatalog())
        indexes = SimpleItem()

        indexes = indexes.__of__(catalog)
        index = index.__of__(indexes)

        return index

    def _makeConflicted(self):
        from ZODB.POSException import ConflictError

        class Conflicted:

            def __str__(self):
                return 'Conflicted'
            __repr__ = __str__

            def __getattr__(self, id, default=object()):
                raise ConflictError('testing')

        return Conflicted()

    def test_empty(self):
        unindex = self._makeOne(id='empty')
        self.assertEqual(unindex.indexed_attrs, ['empty'])

    def test_removeForwardIndexEntry_with_ConflictError(self):
        from ZODB.POSException import ConflictError
        unindex = self._makeOne(id='conflicted')
        unindex._index['conflicts'] = self._makeConflicted()
        self.assertRaises(ConflictError, unindex.removeForwardIndexEntry,
                          'conflicts', 42)

    def test_get_object_datum(self):
        from Products.PluginIndexes.common.UnIndex import _marker
        idx = self._makeOne('interesting')

        dummy = object()
        self.assertEquals(idx._get_object_datum(dummy, 'interesting'), _marker)

        class DummyContent2(object):
            interesting = 'GOT IT'
        dummy = DummyContent2()
        self.assertEquals(idx._get_object_datum(dummy, 'interesting'),
                          'GOT IT')

        class DummyContent3(object):
            exc = None

            def interesting(self):
                if self.exc:
                    raise self.exc
                return 'GOT IT'
        dummy = DummyContent3()
        self.assertEquals(idx._get_object_datum(dummy, 'interesting'),
                          'GOT IT')

        dummy.exc = AttributeError
        self.assertEquals(idx._get_object_datum(dummy, 'interesting'), _marker)

        dummy.exc = TypeError
        self.assertEquals(idx._get_object_datum(dummy, 'interesting'), _marker)

    def test_cache(self):
        idx = self._makeOne(id='foo')
        idx.query_options = ('query', 'range', 'not', 'operator')

        def testQuery(record, expect=1):
            cache = idx.getRequestCache()
            cache.clear()

            # First query
            res1 = idx._apply_index(record)

            # Cache set?
            self.assertEqual(cache._sets, expect)

            # Cache miss?
            self.assertEqual(cache._misses, expect)

            # Second Query
            res2 = idx._apply_index(record)

            # Cache hit?
            self.assertEqual(cache._hits, expect)

            # Check if result of second query is equal to first query
            result = difference(res1[0], res2[0])
            self.assertEqual(len(result), 0)

        # Dummy tests, result is always empty.
        # TODO: Sophisticated tests have to be placed on tests
        # of inherited classes (FieldIndex, KeywordIndex etc.)
        #
        # 'or' operator
        record = {'foo': {'query': ['e', 'f'], 'operator': 'or'}}
        testQuery(record)

        # 'and' operator (currently not supported)
        record = {'foo': {'query': ['e', 'f'], 'operator': 'and'}}
        testQuery(record)

        # 'range' option
        record = {'foo': {'query': ('abc', 'abcd'), 'range': 'min:max'}}
        testQuery(record)

        # 'not' option
        record = {'foo': {'query': ['a', 'ab'], 'not': 'a'}}
        testQuery(record)

    def test_getCounter(self):
        index = self._makeOne('counter')

        self.assertEqual(index.getCounter(), 0)

        class Dummy(object):
            id = 1
            counter = 'counter'

        obj = Dummy()
        index.index_object(obj.id, obj)
        self.assertEqual(index.getCounter(), 1)

        index.unindex_object(obj.id)
        self.assertEqual(index.getCounter(), 2)

        # unknown id
        index.unindex_object(1234)
        self.assertEqual(index.getCounter(), 2)

        # clear changes the index
        index.clear()
        self.assertEqual(index.getCounter(), 3)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import operator
import unittest

from BTrees.IIBTree import IISet
from OFS.SimpleItem import SimpleItem
from Testing.makerequest import makerequest


class Dummy(object):

    def __init__(self, name, start, stop):
        self._name = name
        self._start = start
        self._stop = stop

    def name(self):
        return self._name

    def start(self):
        return self._start

    def stop(self):
        return self._stop

    def datum(self):
        return (self._start, self._stop)


dummies = [Dummy('a', None, None),
           Dummy('b', None, None),
           Dummy('c', 0, None),
           Dummy('d', 10, None),
           Dummy('e', None, 4),
           Dummy('f', None, 11),
           Dummy('g', 0, 11),
           Dummy('h', 2, 9),
           ]


def matchingDummies(value):
    result = []
    for dummy in dummies:
        if ((dummy.start() is None or dummy.start() <= value) and
                (dummy.stop() is None or dummy.stop() >= value)):
            result.append(dummy)
    return result


class DateRangeIndexTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.PluginIndexes.DateRangeIndex.DateRangeIndex \
            import DateRangeIndex
        return DateRangeIndex

    def _makeOne(self, id, since_field=None, until_field=None, caller=None,
                 extra=None):
        klass = self._getTargetClass()
        return klass(id, since_field, until_field, caller, extra)

    def test_interfaces(self):
        from Products.PluginIndexes.interfaces import IDateRangeIndex
        from Products.PluginIndexes.interfaces import IPluggableIndex
        from Products.PluginIndexes.interfaces import ISortIndex
        from Products.PluginIndexes.interfaces import IUniqueValueIndex
        from zope.interface.verify import verifyClass

        verifyClass(IDateRangeIndex, self._getTargetClass())
        verifyClass(IPluggableIndex, self._getTargetClass())
        verifyClass(ISortIndex, self._getTargetClass())
        verifyClass(IUniqueValueIndex, self._getTargetClass())

    def test_empty(self):
        empty = self._makeOne('empty')

        self.assertTrue(empty.getEntryForObject(1234) is None)
        empty.unindex_object(1234)  # shouldn't throw

        self.assertFalse(list(empty.uniqueValues('foo')))
        self.assertFalse(list(empty.uniqueValues('foo', withLengths=True)))
        self.assertTrue(empty._apply_index({'zed': 12345}) is None)

        result, used = empty._apply_index({'empty': 12345})
        self.assertFalse(result)
        self.assertEqual(used, (None, None))

    def test_retrieval(self):
        index = self._makeOne('work', 'start', 'stop')

        for i in range(len(dummies)):
            index.index_object(i, dummies[i])

        for i in range(len(dummies)):
            self.assertEqual(index.getEntryForObject(i), dummies[i].datum())

        for value in range(-1, 15):
            matches = matchingDummies(value)
            results, used = index._apply_index({'work': value})
            self.assertEqual(used, ('start', 'stop'))
            self.assertEqual(len(matches), len(results))

            matches = sorted(matches, key=operator.methodcaller('name'))

            for result, match in map(None, results, matches):
                self.assertEqual(index.getEntryForObject(result),
                                 match.datum())

    def test_longdates(self):
        too_large = 2 ** 31
        too_small = -2 ** 31
        index = self._makeOne('work', 'start', 'stop')
        bad = Dummy('bad', too_large, too_large)
        self.assertRaises(OverflowError, index.index_object, 0, bad)
        bad = Dummy('bad', too_small, too_small)
        self.assertRaises(OverflowError, index.index_object, 0, bad)

    def test_floor_date(self):
        index = self._makeOne('work', 'start', 'stop')
        floor = index.floor_value - 1
        bad = Dummy('bad', floor, None)
        index.index_object(0, bad)
        self.assertTrue(0 in index._always.keys())

    def test_ceiling_date(self):
        index = self._makeOne('work', 'start', 'stop')
        ceiling = index.ceiling_value + 1
        bad = Dummy('bad', None, ceiling)
        index.index_object(1, bad)
        self.assertTrue(1 in index._always.keys())

    def test_datetime(self):
        from datetime import datetime
        from DateTime.DateTime import DateTime
        from Products.PluginIndexes.DateIndex.tests import _getEastern
        before = datetime(2009, 7, 11, 0, 0, tzinfo=_getEastern())
        start = datetime(2009, 7, 13, 5, 15, tzinfo=_getEastern())
        between = datetime(2009, 7, 13, 5, 45, tzinfo=_getEastern())
        stop = datetime(2009, 7, 13, 6, 30, tzinfo=_getEastern())
        after = datetime(2009, 7, 14, 0, 0, tzinfo=_getEastern())

        dummy = Dummy('test', start, stop)
        index = self._makeOne('work', 'start', 'stop')
        index.index_object(0, dummy)

        self.assertEqual(index.getEntryForObject(0),
                         (DateTime(start).millis() / 60000,
                          DateTime(stop).millis() / 60000))

        results, used = index._apply_index({'work': before})
        self.assertEqual(len(results), 0)

        results, used = index._apply_index({'work': start})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': between})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': stop})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': after})
        self.assertEqual(len(results), 0)

    def test_datetime_naive_timezone(self):
        from datetime import datetime
        from DateTime.DateTime import DateTime
        from Products.PluginIndexes.DateIndex.DateIndex import Local
        before = datetime(2009, 7, 11, 0, 0)
        start = datetime(2009, 7, 13, 5, 15)
        start_local = datetime(2009, 7, 13, 5, 15, tzinfo=Local)
        between = datetime(2009, 7, 13, 5, 45)
        stop = datetime(2009, 7, 13, 6, 30)
        stop_local = datetime(2009, 7, 13, 6, 30, tzinfo=Local)
        after = datetime(2009, 7, 14, 0, 0)

        dummy = Dummy('test', start, stop)
        index = self._makeOne('work', 'start', 'stop')
        index.index_object(0, dummy)

        self.assertEqual(index.getEntryForObject(0),
                         (DateTime(start_local).millis() / 60000,
                          DateTime(stop_local).millis() / 60000))

        results, used = index._apply_index({'work': before})
        self.assertEqual(len(results), 0)

        results, used = index._apply_index({'work': start})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': between})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': stop})
        self.assertEqual(len(results), 1)

        results, used = index._apply_index({'work': after})
        self.assertEqual(len(results), 0)

    def test_resultset(self):
        index = self._makeOne('work', 'start', 'stop')
        for i in range(len(dummies)):
            index.index_object(i, dummies[i])

        results, used = index._apply_index({'work': 20})
        self.assertEqual(set(results), set([0, 1, 2, 3]))

        # a resultset with everything doesn't actually limit
        results, used = index._apply_index(
            {'work': 20},
            resultset=IISet(range(len(dummies))))
        self.assertEqual(set(results), set([0, 1, 2, 3]))

        # a small resultset limits
        results, used = index._apply_index(
            {'work': 20},
            resultset=IISet([1, 2]))
        self.assertEqual(set(results), set([1, 2]))

        # the specified value is included
        results, used = index._apply_index({'work': 11})
        self.assertEqual(set(results), set([0, 1, 2, 3, 5, 6]))

        # also for _since_only
        results, used = index._apply_index({'work': 10})
        self.assertEqual(set(results), set([0, 1, 2, 3, 5, 6]))

        # the specified value is included with a large resultset
        results, used = index._apply_index(
            {'work': 11},
            resultset=IISet(range(len(dummies))))
        self.assertEqual(set(results), set([0, 1, 2, 3, 5, 6]))

        # this also works for _since_only
        results, used = index._apply_index(
            {'work': 10},
            resultset=IISet(range(len(dummies))))
        self.assertEqual(set(results), set([0, 1, 2, 3, 5, 6]))

        # the specified value is included with a small resultset
        results, used = index._apply_index(
            {'work': 11},
            resultset=IISet([0, 5, 7]))
        self.assertEqual(set(results), set([0, 5]))

    def test_getCounter(self):
        index = self._makeOne('work', 'start', 'stop')
        self.assertEqual(index.getCounter(), 0)

        obj = dummies[0]
        index.index_object(1, obj)
        self.assertEqual(index.getCounter(), 1)

        index.unindex_object(1)
        self.assertEqual(index.getCounter(), 2)

        # unknown id
        index.unindex_object(1234)
        self.assertEqual(index.getCounter(), 2)

        # clear is a change
        index.clear()
        self.assertEqual(index.getCounter(), 3)


class DateRangeIndexCacheTests(DateRangeIndexTests):

    def _makeOne(self, id, since_field=None, until_field=None,
                 caller=None, extra=None):
        index = super(DateRangeIndexCacheTests, self)._makeOne(
            id, since_field, until_field, caller, extra)

        class DummyZCatalog(SimpleItem):
            id = 'DummyZCatalog'

        # Build pseudo catalog and REQUEST environment
        catalog = makerequest(DummyZCatalog())
        indexes = SimpleItem()

        indexes = indexes.__of__(catalog)
        index = index.__of__(indexes)

        return index

    def test_cache(self):
        index = self._makeOne('work', 'start', 'stop')
        for i in range(len(dummies)):
            index.index_object(i, dummies[i])

        cache = index.getRequestCache()

        # without resultset
        cache.clear()

        # first call
        results_1, used = index._apply_index({'work': 20})
        self.assertEqual(set(results_1), set([0, 1, 2, 3]))
        self.assertEqual(cache._hits, 0)
        self.assertEqual(cache._sets, 1)
        self.assertEqual(cache._misses, 1)

        # second call
        results_2, used = index._apply_index({'work': 20})
        self.assertEqual(set(results_1), set(results_2))
        self.assertEqual(cache._hits, 1)

        # with resultset
        cache.clear()

        # first call
        results_1, used = index._apply_index(
            {'work': 20},
            resultset=IISet(range(len(dummies))))
        self.assertEqual(set(results_1), set([0, 1, 2, 3]))
        self.assertEqual(cache._hits, 0)
        self.assertEqual(cache._sets, 1)
        self.assertEqual(cache._misses, 1)

        # second call
        results_2, used = index._apply_index(
            {'work': 20},
            resultset=IISet(range(len(dummies))))
        self.assertEqual(set(results_1), set(results_2))
        self.assertEqual(cache._hits, 1)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from BTrees.IIBTree import difference
from BTrees.IIBTree import IIBucket
from BTrees.IIBTree import weightedIntersection
from BTrees.IIBTree import weightedUnion
from BTrees.OOBTree import OOSet
from BTrees.OOBTree import union


class ResultList:

    def __init__(self, d, words, index):
        self._index = index

        if type(words) is not OOSet:
            words = OOSet(words)
        self._words = words

        if (type(d) is tuple):
            d = IIBucket((d,))
        elif type(d) is not IIBucket:
            d = IIBucket(d)

        self._dict = d
        self.__getitem__ = d.__getitem__
        try:
            self.__nonzero__ = d.__nonzero__
        except Exception:
            pass
        self.get = d.get

    def __nonzero__(self):
        return not not self._dict

    def bucket(self):
        return self._dict

    def keys(self):
        return self._dict.keys()

    def has_key(self, key):
        return key in self._dict

    def items(self):
        return self._dict.items()

    def __and__(self, x):
        return self.__class__(
            weightedIntersection(self._dict, x._dict),
            union(self._words, x._words),
            self._index,
        )

    def and_not(self, x):
        return self.__class__(
            difference(self._dict, x._dict),
            self._words,
            self._index,
        )

    def __or__(self, x):
        return self.__class__(
            weightedUnion(self._dict, x._dict),
            union(self._words, x._words),
            self._index,
        )

    def near(self, x):
        result = IIBucket()
        dict = self._dict
        xdict = x._dict
        xhas = xdict.has_key
        positions = self._index.positions
        for id, score in dict.items():
            if not xhas(id):
                continue
            p = (map(lambda i: (i, 0), positions(id, self._words)) +
                 map(lambda i: (i, 1), positions(id, x._words)))
            p.sort()
            d = lp = 9999
            li = None
            lsrc = None
            for i, src in p:
                if i is not li and src is not lsrc and li is not None:
                    d = min(d, i - li)
                li = i
                lsrc = src
            if d == lp:
                score = min(score, xdict[id])  # synonyms
            else:
                score = (score + xdict[id]) / d
            result[id] = score

        return self.__class__(
            result, union(self._words, x._words), self._index)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
#############################################################################

import types


def safe_callable(ob):
    # Works with ExtensionClasses and Acquisition.
    if hasattr(ob, '__class__'):
        return hasattr(ob, '__call__') or isinstance(ob, types.ClassType)
    else:
        return callable(ob)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from types import InstanceType

from DateTime.DateTime import DateTime

_marker = []


class IndexRequestParseError(Exception):
    pass


class parseIndexRequest:
    """
    This class provides functionality to hide the internals of a request
    send from the Catalog/ZCatalog to an index._apply_index() method.

    The class understands the following type of parameters:

    - old-style parameters where the query for an index as value inside
      the request directory where the index name is the name of the key.

    - dictionary-style parameters specify a query for an index as
      an entry in the request dictionary where the key corresponds to the
      name of the index and the key is a dictionary with the parameters
      passed to the index.

      Allowed keys of the parameter dictionary:

      'query'  - contains the query (either string, list or tuple) (required)

      other parameters depend on the the index

   - record-style parameters specify a query for an index as instance
     of the Record class. This happens usually when parameters from a
     web form use the "record" type e.g.
           <input type="text" name="path.query:record:string">.
     All restrictions of the dictionary-style parameters apply to the
     record-style parameters
    """

    ParserException = IndexRequestParseError

    def __init__(self, request, iid, options=[]):
        """ parse a request  from the ZPublisher and return a uniform
        datastructure back to the _apply_index() method of the index

          request -- the request dictionary send from the ZPublisher
          iid     -- Id of index
          options -- a list of options the index is interested in
        """

        self.id = iid
        if iid not in request:
            self.keys = None
            return

        param = request[iid]
        keys = None

        if isinstance(param, InstanceType) and not isinstance(param, DateTime):
            """ query is of type record """

            record = param

            if not hasattr(record, 'query'):
                raise self.ParserException(
                    "record for '%s' *must* contain a "
                    "'query' attribute" % self.id)
            keys = record.query

            if isinstance(keys, str):
                keys = [keys.strip()]

            for op in options:
                if op == "query":
                    continue

                if hasattr(record, op):
                    setattr(self, op, getattr(record, op))

        elif isinstance(param, dict):
            """ query is a dictionary containing all parameters """

            query = param.get("query", ())
            if isinstance(query, (tuple, list)):
                keys = query
            else:
                keys = [query]

            for op in options:
                if op == "query":
                    continue

                if op in param:
                    setattr(self, op, param[op])

        else:
            """ query is tuple, list, string, number, or something else """

            if isinstance(param, (tuple, list)):
                keys = param
            else:
                keys = [param]

            for op in options:
                field = iid + "_" + op
                if field in request:
                    setattr(self, op, request[field])

        self.keys = keys
        not_value = getattr(self, 'not', None)
        if not_value is not None:
            if not isinstance(not_value, (tuple, list)):
                not_value = [not_value]
                setattr(self, 'not', not_value)

    def get(self, k, default_v=None):
        if hasattr(self, k):
            v = getattr(self, k)
            if v != '':
                return v
        return default_v


class RequestCache(dict):

    # stats info needed for testing
    _hits = 0
    _misses = 0
    _sets = 0

    def get(self, key, default=None):
        value = super(RequestCache, self).get(key, _marker)

        if value is _marker:
            self._misses += 1
            return default

        self._hits += 1
        return value

    def __getitem__(self, key):
        try:
            value = super(RequestCache, self).__getitem__(key)
        except KeyError as e:
            self._misses += 1
            raise e

        self._hits += 1
        return value

    def __setitem__(self, key, value):
        super(RequestCache, self).__setitem__(key, value)
        self._sets += 1

    def clear(self):
        super(RequestCache, self).clear()
        self._hits = 0
        self._misses = 0
        self._sets = 0

    def stats(self):
        stats = {'hits': self._hits,
                 'misses': self._misses,
                 'sets': self._sets}
        return stats

    def __str__(self):
        return "<RequestCache %s items (hits: %s, misses: %s, sets: %s)>" % \
            (len(self), self._hits, self._misses, self._sets)
